/** Interface VRP simplifie pour ALTA. Initialisation par Madeleine.
 */

#include "VRP-internals.h"
#include "VRP-ALTA.h"
#include "pm2_common.h"
#include <sys/poll.h>
#include <assert.h>

struct vrp_alta_s
{
  p_mad_madeleine_t madeleine;
  p_mad_channel_t channel;
  int size;
  vrp_incoming_t*incomings;
  vrp_outgoing_t*outgoings;
  void**posted_buffer;
  int posted_elm_size;
  int posted_offset;
  int*posted_n_elm;
  struct vrp_alta_loss_feedback_s*posted_losses;
};
typedef struct vrp_alta_s*vrp_alta_t;

static struct vrp_alta_s vrp_alta =
  {
    .madeleine = NULL,
    .channel   = NULL,
    .size      = 0,
    .incomings = NULL,
    .outgoings = NULL
  };


static void vrp_alta_frame_handler(vrp_in_buffer_t buffer);

/* ********************************************************* */

static void vrp_alta_resize(int rank)
{
  const int size = rank + 1;
  if(size > vrp_alta.size)
    {
      vrp_alta.incomings = realloc(vrp_alta.incomings, size * sizeof(vrp_incoming_t));
      vrp_alta.outgoings = realloc(vrp_alta.outgoings, size * sizeof(vrp_outgoing_t));
      memset(vrp_alta.incomings + vrp_alta.size, 0, (size-vrp_alta.size)*sizeof(vrp_incoming_t));
      memset(vrp_alta.outgoings + vrp_alta.size, 0, (size-vrp_alta.size)*sizeof(vrp_outgoing_t));
      vrp_alta.size = size;
    }
}

void vrp_alta_set_incoming(vrp_incoming_t incoming, int rank)
{
  vrp_alta_resize(rank);
  vrp_alta.incomings[rank] = incoming;
}

void vrp_alta_set_outgoing(vrp_outgoing_t outgoing, int rank)
{
  vrp_alta_resize(rank);
  vrp_alta.outgoings[rank] = outgoing;
}

/* ********************************************************* */

void vrp_alta_connect(void)
{
  int i;
  for(i = 0; i < vrp_alta.size; i++)
    {
      vrp_outgoing_t outgoing = vrp_alta.outgoings[i];
      if(outgoing)
	vrp_outgoing_connect(outgoing);
    }
}

void vrp_alta_send(int dest_rank,
		   const void*data, int offset, int size_elm, int n_elm,
		   const struct vrp_alta_loss_parameters_s*loss_params)
{
  vrp_outgoing_t outgoing = vrp_alta.outgoings[dest_rank];
  vrp_buffer_t buffer = vrp_buffer_construct((void*)data, offset + size_elm * n_elm);
  if(offset > 0)
    {
      vrp_buffer_add_critical(buffer, (void*)data, offset);
    }
  vrp_buffer_set_loss_rate(buffer, loss_params->max_loss_rate, loss_params->max_cons_loss);
  if(loss_params->min_refresh_rate != -1)
    {
      vrp_outgoing_set_min_refresh_rate(outgoing, loss_params->min_refresh_rate);
    }
  vrp_outgoing_send_frame(outgoing, buffer);
  vrp_buffer_destroy(buffer);
}

void vrp_alta_recv(void**buffer, int offset, int size_elm, int*n_elm,
		   vrp_alta_loss_feedback_t losses)
{
  int i, j = 0;
  int nfds = 0;
  int done = 0;
  struct pollfd*pi = vrp_malloc(sizeof(struct pollfd)*vrp_alta.size);
  struct pollfd*po = vrp_malloc(sizeof(struct pollfd)*vrp_alta.size);
  for(i = 0; i < vrp_alta.size; i++)
    {
      vrp_incoming_t incoming = vrp_alta.incomings[i];
      if(incoming)
	{
	  pi[j] = (struct pollfd)
	    {
	      .fd = vrp_incoming_fd(incoming),
	      .events = POLLIN
	    };
	  /*
	  vrp_outgoing_t outgoing = vrp_alta.outgoings[i];
	  po[j] = (struct pollfd)
	    {
	      .fd = vrp_outgoing_fd(outgoing),
	      .events = POLLIN
	    };
	  */
	  j++;
	}
    }
  nfds = j;
  while(!done)
    {
      /*
      marcel_extlib_protect();
      int rc1 = poll(po, nfds, 0);
      marcel_extlib_unprotect();
      if(rc1 > 0)
	{
	  for(i = 0, j = 0; i < nfds; i++, j++)
	    {
	      if(!vrp_alta.outgoings[j])
		j++;
	      if(po[i].revents & POLLIN)
		{
		  vrp_outgoing_t outgoing = vrp_alta.outgoings[j];
		  vrp_outgoing_read(outgoing);
		}
	    }
	}
      */
      marcel_extlib_protect();
      int rc2 = poll(pi, nfds, 0);
      marcel_extlib_unprotect();
      if(rc2 > 0)
	{
	  for(i = 0, j = 0; i < nfds; i++, j++)
	    {
	      if(!vrp_alta.outgoings[j])
		j++;
	      if(pi[i].revents & POLLIN)
		{
		  vrp_incoming_t incoming = vrp_alta.incomings[j];
		  vrp_incoming_set_handler(incoming, &vrp_alta_frame_handler);
		  vrp_alta.posted_buffer = buffer;
		  vrp_alta.posted_offset = offset;
		  vrp_alta.posted_elm_size = size_elm;
		  vrp_alta.posted_n_elm = n_elm;
		  vrp_alta.posted_losses = losses;
		  vrp_incoming_read(incoming);
		  if(vrp_alta.posted_buffer == NULL)
		    {
		      done = 1;
		    }
		  vrp_alta.posted_buffer = NULL;
		}
	    }
	}
      if(rc2 <= 0)
	{
	  marcel_delay(10);
	}
    }
  vrp_free(pi);
  vrp_free(po);
}

static void vrp_alta_frame_handler(vrp_in_buffer_t buffer)
{
  assert(vrp_alta.posted_buffer != NULL);
  void*data = vrp_malloc(buffer->size);
  struct vrp_alta_loss_feedback_s*losses = vrp_alta.posted_losses;
  memcpy(data, buffer->data, buffer->size);
  *vrp_alta.posted_buffer = data;
  vrp_alta.posted_buffer = NULL;
  const int n = (buffer->size - vrp_alta.posted_offset) / vrp_alta.posted_elm_size;
  *vrp_alta.posted_n_elm = n;
  /* fill feedback losses */
  losses->losses_array = vrp_malloc(n);
  losses->size = n;
  memset(losses->losses_array, 0, n);
  int i;
  for(i = 0; i < buffer->loss_num; i++)
    {
      int l = buffer->lost[i] - data;
      int j = 0; /* j+l : curseur dans 'data' */
      int k = l / vrp_alta.posted_elm_size; /* curseur dans 'losses_array' */
      while(j <= buffer->loss_granularity && (j+l) < buffer->size)
	{
	  losses->losses_array[k] = 1;
	  j++;
	  k += vrp_alta.posted_elm_size;
	}
    }

}
